﻿var fs = require('fs');
var Promise = require('promise');
var jsMapper = require("./js_diff/index.js").javascriptHandler;
var mapSingleFileInMemory = require("./js_diff/index.js").mapSingleFileInMemory;
var consoleFileLogger = require("./sl-logger.js").consoleFileLogger;
var version = require("../package.json").version;
var slNodeCover = require('sl-node-cover');
var SlNeedle = require('./sl-needle.js').SlNeedle;
var NewDiffService = require("./new-diff-service.js").NewDiffService;
var md5 = require('md5');

function BuildDiffProcess(cfg, diffService, sourceConrolProviders, parsedDependencies, logger) {

    var log = logger ? logger.child({
        className: "BuildDiffProcess"
    }) : consoleFileLogger("BuildDiffProcess");

    function createBuildDiffPromiseChain(buildArguments) {
        var handlerParams = {
            buildArguments: buildArguments,
            cfg: cfg,
            finalOutput: {
                "meta": {
                    "generated": new Date().valueOf(),
                    "build": buildArguments.build,
                    "commit": buildArguments.commit,
                    "customerId": cfg.customerId,
                    "appName": buildArguments.appname,
                    //"technologies": "nodejs",
                    "technology": "nodejs",
                    "branch": buildArguments.branch,
                    "environment": {
                        agentVersion: version
                    },
                    "authors": buildArguments.author,
                    "logsUrl": buildArguments.logsUrl,
                    "jobName": buildArguments.jobName
                },
                files: [],
                dependencies: parsedDependencies || []
            }
        };

        var scmInstance;

        if (buildArguments.workspacepath) {
            var scmType = sourceConrolProviders[buildArguments.scm];
            if (scmType) {
                scmInstance = new scmType(cfg, buildArguments.workspacePath);
                handlerParams.finalOutput.meta.scm = buildArguments.scm;
            }
        }

        if (!buildArguments.commit && scmInstance && !scmInstance.detectCommitVersion) {
            return Promise.reject("'commit' command-line argument is required, since automatic detection by the " + buildArguments.scm + " SCM is not supported");
        }

        var promiseChain = Promise.resolve();

        if (buildArguments.workspacepath) {

            //2. If there is a mapping, get the diffs
            if (scmInstance) {
                //2.1 Get repository url from the SCM
                promiseChain = promiseChain.then(function (mapping) {
                    return getRepositoryUrl(handlerParams, scmInstance);
                });

                //2.2 Get branch history from the SCM
                promiseChain = promiseChain.then(function (mapping) {
                    return loadBranchHistory(handlerParams, scmInstance);
                });

                //2.3 if commit was not specified, automatically detect it
                if (!buildArguments.commit) {
                    promiseChain = promiseChain.then(function () {
                        return scmInstance.detectCommitVersion(handlerParams, log).then(function (commit) {
                            if (commit) {
                                handlerParams.finalOutput.meta.commit = commit;
                            }
                        });
                    });
                };
            }

            //3. Call JS mapping handler
            promiseChain = promiseChain.then(function () {
                return jsMapper(handlerParams, log);
            });
        };

        //4. Submit the final output to the server
        promiseChain = promiseChain.then(function () {
            return diffService.submitMapping(handlerParams.finalOutput);
        });

        //5. Create instrumented browser files.
        if (buildArguments.instrumentForBrowsers) {
            promiseChain = promiseChain.then(function () {
                var promise = new Promise(function (resolve, reject) {
                    slNodeCover.utils.runInstrumentation({
                        appName: buildArguments.appname,
                        build: buildArguments.build,
                        customerId: cfg.customerId,
                        branch: buildArguments.branch,
                        sourceRoot: buildArguments.workspacepath,
                        files: handlerParams.processedFiles, //Will be filled by jsMapper in step 3
                        outputPath: buildArguments.outputpath,
                        buildsessionid: buildArguments.buildsessionid,
                        server: handlerParams.cfg.server,
                        token: cfg.token,
                        instrumentationType: "browser"
                    }, function (err) {
                        if (err)
                            reject(err);
                        else
                            return resolve();
                    });
                });
                return promise;
            });
        }

        promiseChain = promiseChain.then(function () {
            log.info("Build agent finished");
            return Promise.resolve();
        }).catch(function (err) {
            log.info("Build agent error: " + err);
            return Promise.reject(err);
        });

        return promiseChain;
    }

    this.run = function run(buildArguments) {
        return createBuildDiffPromiseChain(buildArguments);
    }

    function loadBranchHistory(handlerParams, scmInstance) {
        return scmInstance.getBranchHistory(handlerParams, log).then(function (gitLog) {
            if (gitLog && gitLog.commitHistory && gitLog.commitHistory.length && gitLog.commitLog && gitLog.commitLog.length) {
                log.info('Branch history loaded (' + gitLog.length + ' commits)');
                handlerParams.finalOutput.meta.history = gitLog.commitHistory;
                handlerParams.finalOutput.meta.commitLog = gitLog.commitLog;
            }
            return gitLog;
        });
    }

    function getRepositoryUrl(handlerParams, scmInstance) {
        return scmInstance.getRepositoryUrl(handlerParams, log).then(function (repositoryUrl) {
            if (repositoryUrl) {
                log.info('Repository Url loaded (' + repositoryUrl + ')');
                handlerParams.finalOutput.meta.repositoryUrl = repositoryUrl;
            }
            return repositoryUrl;
        });
    }
}

function InstrumentingProxyTransformingProcess(diffService) {
    this.run = function run(options, logger) {
        var logger = logger || new consoleFileLogger("build-scanner-proxy");
        var needle = new SlNeedle(options.token);
        var diffService = new NewDiffService({
            proxy: options.proxy,
            isValidToken: options.isValidToken,
            server: options.server
        }, needle, logger);

        var fileSig = mapSingleFileInMemory({
            filename: options.filename,
            fullFilename: options.originalUrl,
            buffer: options.originalFileContent,
            usebranchcoverage: true
        }, logger);

        var moduleName = md5(options.originalUrl); //Uncomment this to include hash in module name, but this ruin trends since modules will never be the same: /* + "#" + fileSig.hash */

        var finalOutput = {
            "meta": {
                "generated": new Date().valueOf(),
                "build": options.buildName,
                "commit": "",
                "customerId": options.customerId,
                "appName": options.appName,
                "moduleName": moduleName,
                "technology": "nodejs",
                "branch": options.branchName,
                "environment": {
                    agentVersion: version
                }
            },
            files: [fileSig],
            dependencies: []
        };

        var promiseChain = Promise.resolve();
        promiseChain = promiseChain.then(function () {
            logger.info("Submitting mapping...");
            return diffService.submitMapping(finalOutput);
        }).then(function () {
            var promise = new Promise(function (resolve, reject) {
                var inMemoryFile = {
                    filename: options.filename,
                    content: options.originalFileContent,
                    instrumented: null
                };

                slNodeCover.utils.runInstrumentation({
                    appName: options.appName,
                    build: options.buildName,
                    customerId: options.customerId,
                    branch: options.branchName,
                    files: [],
                    inMemoryFile: inMemoryFile,
                    additionalFilesToCopy: [],
                    outputPath: null,
                    buildsessionid: options.buildsessionid,
                    server: options.server,
                    token: options.token,
                    instrumentationType: "browser"
                }, function (err) {
                    if (err) {
                        logger.error("Error", err);
                        reject(err);
                    } else {
                        logger.info("Done (success)");
                        return resolve(inMemoryFile.instrumented);
                    }
                });
            });
            return promise;
        });

        return promiseChain;
    }
}

module.exports.InstrumentingProxyTransformingProcess = InstrumentingProxyTransformingProcess;
module.exports.BuildDiffProcess = BuildDiffProcess;
module.exports.DiffService = require("./new-diff-service.js").NewDiffService;
module.exports.sourceControlProviders = require("./source-control-providers.js");
