﻿var glob = require("glob"),
    Promise = require('promise'),
    path = require('path'),
    fs = require('fs'),
    fileSignature = require('./file-signature.js'),
    methodKeyCreator = require('./method-key.js'),
    uuid = require('uuid'),
    //statuses = require('../statuses.js'),
    util = require('util'),
    parser = require("gitignore-parser"),
    sourceMap = require("source-map"),
    utils = require('../utils'),
    consoleFileLogger = require("../sl-logger.js").consoleFileLogger;

function ActualFile(filename, fullFilename) {
    this.getFilename = function ActualFile$getFilename(){ return filename; }
    this.getFullFilename = function ActualFile$getFullFilename(){ return fullFilename; }
    this.getContent = function ActualFile$getContent(){
        var content = fs.readFileSync(fullFilename, 'utf-8');
        return content;
    }
    this.readSourceMaps = function ActualFile$readSourceMaps(logger){
        return readSourceMaps(fullFilename, logger);
    } 
}

function InMemoryFile(filename, originalFilename, buffer){
    this.getFilename = function InMemoryFile$getFilename(){ return filename; }
    this.getFullFilename = function InMemoryFile$getFullFilename(){ return originalFilename; }
    this.getContent = function InMemoryFile$getContent(){
        return buffer;
    }
    this.readSourceMaps = function InMemoryFile$readSourceMaps(logger){
        return null;
    }
}

function generateNewMappingForFile(fileObject, isBranchCoverage, logger) {
    var filename = fileObject.getFilename();
    filename = utils.adjustPathSlashes(filename);
    var fileSig = {
        filename: filename,
        methods: []
    };
    
    var oldContent = fileObject.getContent();
    var sourceMaps = fileObject.readSourceMaps(logger);
    var sig = fileSignature(filename, oldContent, sourceMaps, isBranchCoverage, logger);
    if (!sig || sig.error) {
        return null;
    }
    fileSig.hash = sig.module.hash;

    var mapping = {};
    sig.methods.forEach(function (m) {
        var wholeKey = [methodKeyCreator(m.start, m.name, m.isAnonymous), m.hash, new Date().valueOf()].join('|');
        var newId = uuid(wholeKey);
        fileSig.methods.push({
            id: newId,
            isAnonymous: m.isAnonymous,
            hash: m.hash,
            sigHash: m.sigHash,
            position: m.start,
            endPosition: m.end,
            name: m.name,
            uniqueName: m.uniqueName,
            filename: m.originalFilename,
            originalFilename: m.filename,
            originalStart: m.originalStart,
            originalEnd: m.originalEnd,
            originalName: m.originalName,
            branches: m.branches
        })
    });
    return fileSig;
}

function readSourceMaps(fullFilename, logger){
	if (!fullFilename) return;
	var sourceMapsFilename = fullFilename+'.map';
	var contents,rawSourceMapJsonData,consumer;
    var log = logger? logger.child({methodName: "readSourceMaps"}) : consoleFileLogger("readSourceMaps");
	try{
		contents = fs.readFileSync(sourceMapsFilename).toString();		
	}
	catch(e) {
		//Sourcemap file not found
		return null;		
	}

    try {
        rawSourceMapJsonData = JSON.parse(contents);
    } catch (e) {
        log.error('Ignoring invalid source map: ' + sourceMapsFilename);
        return null;
    }

    try {
        log.debug('Processing source maps from file: ' + sourceMapsFilename);
        consumer = new sourceMap.SourceMapConsumer(rawSourceMapJsonData);
        return consumer;
    } catch (e) {
        log.error('Error processing source maps for file: ' + sourceMapsFilename);
        return null;
    }
}

function findSlIgnore(workspacePath, lookInParentFolders, logger) {
    var searchedPaths = {};
    do {
        var slIgnoreFilePath = path.resolve(workspacePath, '.slignore');
        if (searchedPaths[slIgnoreFilePath]) return null; //already searched at

        logger.debug('Looking for .slignore file at ' + slIgnoreFilePath);
        searchedPaths[slIgnoreFilePath] = true;
        try {
            var f = fs.statSync(slIgnoreFilePath);
            if (f.isFile()) return slIgnoreFilePath;
        } catch (e) {
            workspacePath = workspacePath + '/..';
        }
        if (!lookInParentFolders) break;
    } while (workspacePath.length > 1);
    logger.info('.slignore file not found');
    return null;
}

function getSlIgnore(workspacePath, outputPath, lookInParentFolders, logger) {
    var log = logger ? logger.child({
        methodName: "getSlIgnore"
    }) : consoleFileLogger("getSlIgnore");
    
    var pattern;
    try {
        var slIgnoreFilePath = findSlIgnore(workspacePath, lookInParentFolders, logger);
        if (!slIgnoreFilePath) throw new Error("SLIGNORE NOT FOUND");
        pattern = fs.readFileSync(slIgnoreFilePath, 'utf8');
        pattern = addExcludedPath(pattern, outputPath);
        try {
            var compiled = parser.compile(pattern); //may throw if the file is invalid/corrupted
            log.info('Using .slignore file ' + slIgnoreFilePath);
            return compiled;
        } catch(e) {
            log.error('Not using corrupted .slignore file ' + slIgnoreFilePath+". Using default behavior.");
            throw e;
        }
    }
    catch (e) {
        // default pattern - all js files under test directory and it's sub directories.
        pattern = "test/**.js";
        pattern = addExcludedPath(pattern, outputPath);
        return parser.compile(pattern); //should not throw again
    }    
}

function addExcludedPath(pattern, excludedFolder) {
    if (!excludedFolder) return pattern;
    if (excludedFolder && excludedFolder[excludedFolder.length-1] != '/') { //Ensure the excludedFolder ends with / to comply with gitignore
        excludedFolder += '/';
    }
    pattern = pattern + "\n" + excludedFolder;
    return pattern;
}

function mapSingleFileInMemory(singleFileParams, logger) { 
    var useBranchCoverage = singleFileParams.usebranchcoverage;
    var fileObject = new InMemoryFile(singleFileParams.filename, singleFileParams.originalFilename, singleFileParams.buffer);
    var fileSig = generateNewMappingForFile(fileObject, useBranchCoverage, logger);
    return fileSig;
}

function javascriptHandler(handlerParams, logger) {
    var log = logger? logger.child({methodName: "javascriptHandler"}) : consoleFileLogger("javascriptHandler");
    var opts = handlerParams.buildArguments;
    var slIgnore = getSlIgnore(opts.workspacepath, opts.outputpath, !!opts.instrumentForBrowsers, log);
    var finalOutput = handlerParams.finalOutput;

    var p = new Promise(function (resolve, reject) {
        finalOutput.files = finalOutput.files || [];
        var files = [];
        var useSlIgnore = true;
        if (handlerParams.buildArguments.specificFiles)
        {
            files = handlerParams.buildArguments.specificFiles;
            useSlIgnore = false;
        }
        else
        {
            files = glob.sync("**/*.js",
            {
                cwd: opts.workspacepath,
                ignore: ["node_modules/**/*.js"]
            });
        }
        handlerParams.processedFiles = [];        

        files.forEach(function (filename, fileIdx) {
            var fullFilename = path.resolve(opts.workspacepath, filename);

            // If the file is filtered by the .slignore file:
            if(useSlIgnore && slIgnore.denies(filename)) {
                log.debug('Excluded [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);                
                return; // Ignored by .slignore, skip it.
            }
            handlerParams.processedFiles.push(filename);

            log.info('Processing [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);
            
            var useBranchCoverage = handlerParams.cfg.useBranchCoverage || opts.usebranchcoverage;

            var fileObject = new ActualFile(filename, fullFilename);

            var fileSig = generateNewMappingForFile(fileObject, useBranchCoverage, logger);
            //If source maps are used, the methods in fileSig may originate from excluded files
            
            if (!useSlIgnore) {
                var includedMethods = [];
                fileSig.methods.forEach(function(m){
                    if (m.filename) {
                        if (!slIgnore.denies(m.filename)){
                            includedMethods.push(m);
                            delete m.filename;
                        }
                        else {
                            //console.log('Rejecting method from file: '+m.filename);
                        }
                    } else{
                        includedMethods.push(m);
                    }
                });

                fileSig.methods = includedMethods;
            }
            else {
                fileSig.methods.forEach(function(m){ delete m.filename; });
            }

            if (fileSig) finalOutput.files.push(fileSig);
        })
        resolve(true);
    });
    return p;
}

module.exports.javascriptHandler = javascriptHandler;
module.exports.mapSingleFileInMemory = mapSingleFileInMemory;